
from copy import copy
import io
import re
import uuid
import inspect
import sys

from abc import ABCMeta
import warnings

import logging
log = logging.getLogger(__name__)

from partis.utils import (
  cached_property,
  adict,
  odict,
  rdict,
  rlist,
  StringFunction,
  mapping_attrs,
  protected_attr,
  indent_lines,
  split_lines,
  fmt_obj,
  fmt_base_or_type,
  _fmt_class_name,
  fmt_class_name,
  fmt_attr_doc,
  make_dynamic_class_name )

from partis.pyproj import (
  hash_sha256 )

from partis.utils.special import (
  required,
  optional,
  derived )

from .base import (
  schema_errors,
  SchemaError,
  SchemaNameError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaHint,
  Loc,
  PresetValue,
  assert_valid_name,
  assert_valid_path,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_special,
  is_optional,
  is_required,
  is_derived,
  is_similar_value_type,
  is_schema_prim,
  is_schema_declared,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated_class,
  is_evaluated,
  is_valued,
  is_valued_type,
  any_schema )

from .property import (
  SchemaProperty,
  ConstProperty )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def fmt_schema_typename( obj, fmt = _fmt_class_name ):

  if isinstance( obj, SchemaDeclared ):
    obj = obj.schema

  if isinstance( obj, Schema ):
    if isinstance( obj.valued_type, SchemaProxy ):
      obj = obj.valued_type

  if isinstance( obj, SchemaProxy ):
    return fmt(obj)

  elif isinstance( obj, Schema ):
    return f"{fmt(obj.valued_type)} [ {fmt(obj)} ]"

  return obj

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def PassValued(
  val = None,
  schema = None,
  loc = None ):
  """Dummy function that simply returns the original un-modified value

  Note
  ----
  This technically not related to the 'Valued' classes, but provides for the
  ability to 'pass-through' a value using a similar call signature as the 'Valued'
  classes.
  """

  return val

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaRef( ABCMeta ):
  """Base class of all schema-like class types (is or uses a schema)
  """

  #-----------------------------------------------------------------------------
  @property
  def schema( cls ):
    """:class:`SchemaDep <partis.schema_meta.schema.SchemaDep>` : The defined schema.
    """
    raise NotImplementedError(f"Not implemented by {cls}")

  #-----------------------------------------------------------------------------
  @property
  def schema_defined( cls ):
    """bool: True if the schema has been defined for this reference.
    """

    raise NotImplementedError(f"Not implemented by {cls}")

  #-----------------------------------------------------------------------------
  def subclass( cls,
    name = None,
    module = None,
    module_set = None,
    **kwargs ):
    """Generator function for sub-classing the schema class

    Parameters
    ----------
    name : NoneType | str
      Name of subclass.
      If None, will generate a name automatically.
    module : None | str | ModuleType
      Module to use for the subclass.
      If None, the module will be determined from the caller frame.
      If a string, the module must be importable.
    module_set : None | bool
      If True (or None), adds the subclass to the module `module`.
      If False, the return value will be the only reference to the subclass.
    **kwargs :
      Keyword arguments passed to `__new__` and `__init__` of
      :class:`SchemaMeta <partis.schema_meta.schema.SchemaMeta>`.

    Returns
    -------
    subclass : :class:`SchemaMeta <partis.schema_meta.schema.SchemaMeta>`
      New subclass with the current class as the only base class.
    """

    module, name = make_dynamic_class_name(
      default_name = cls.__name__,
      module = module,
      name = name )

    bases = ( cls, )

    namespace = type(cls).__prepare__(
      name,
      bases,
      **kwargs )

    namespace['__module__'] = module.__name__

    sub_cls = type(cls)(
      name,
      bases,
      namespace,
      **kwargs )

    if module_set is None or module_set:

      if hasattr( module, sub_cls.__name__ ):
        warnings.warn(f"Overriding module attribute with class definition: {module}, {sub_cls.__name__}")

      setattr( module, sub_cls.__name__, sub_cls )


    return sub_cls


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaProxy( SchemaRef ):
  """Base class for types that use a schema, by is not itself a schema

  Parameters
  ----------
  name : str
    Class name of type
  bases : list[type]
    Base classes of type
  namespace : dict[str, object]
    Type namespace (class body)
  schema : :class:`SchemaDep <partis.schema_meta.schema.SchemaDep>`
    The defined schema.
  """

  #-----------------------------------------------------------------------------
  def __new__( mcls,
    name,
    bases,
    namespace,
    schema ):

    if schema and not isinstance( schema, SchemaDep ):
      raise SchemaError(
        f"Schema proxy must be a SchemaDep.")


    namespace = {
      **namespace,
      '_p_schema': schema,
      # NOTE: generates docstring on-demand using _get_doc
      '__doc__' : StringFunction() }

    cls = super().__new__(
      mcls,
      name,
      bases,
      namespace )

    cls.__doc__._func = cls._get_doc

    return cls

  #-----------------------------------------------------------------------------
  def __init__( cls,
    name,
    bases,
    namespace,
    schema ):

    super().__init__( name, bases, namespace )

  #-----------------------------------------------------------------------------
  def _get_doc( cls,
    noindex = False,
    depth = 0,
    max_depth = 2 ):
    """Internal method to generate the docstring for the class

    Note
    ----
    The normal `__doc__` attribute should be used to access the generated
    docstring, which caches the result of this method.

    Parameters
    ----------
    noindex : bool
      Whether to add the attribute references to the doc index.

      Used with recursivly including schema docstrings in-place, preventing
      multiple index references to the same attribute, or for keeping the doc
      of 'class' attributes from clashing with any 'instance' attributes with
      the same name in the index.

      .. note::

        Even when True, the docstring will still be included in the generation
        of the documentation, but it won't be referencable.

    depth : int
      Current depth of recursively included docstrings
    max_depth : int
      Maximum depth to recursively included docstrings of related schema classes

    Returns
    -------
    str
    """

    return cls.schema._get_doc(
      noindex = True,
      depth = depth,
      max_depth = max_depth )

  #-----------------------------------------------------------------------------
  @property
  def schema( cls ):
    return cls._p_schema

  #-----------------------------------------------------------------------------
  @property
  def schema_defined( cls ):
    return cls._p_schema.schema_defined

  #-----------------------------------------------------------------------------
  def schema_resolvable( cls, checking = None ):
    return cls._p_schema.schema_resolvable(checking = checking)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaDep( SchemaRef ):
  """Base class of concrete schema types

  Parameters
  ----------
  name : str
    Class name of type
  bases : list[type]
    Base classes of type
  namespace : dict[str, object]
    Type namespace (class body)
  schema_deps : None | list[ :class:`SchemaRef <partis.schema_meta.schema.SchemaRef>` ]
    Other schemas which this schema depends upon to be defined.
  schema_refs : None | list[ :class:`SchemaRef <partis.schema_meta.schema.SchemaRef>` ]
    Other schemas which depend on (defined with) this schema.
  """

  #-----------------------------------------------------------------------------
  def __new__( mcls,
    name,
    bases,
    namespace,
    schema_deps = None,
    schema_refs = None ):

    if schema_deps is None:
      schema_deps = list()

    if schema_refs is None:
      schema_refs = list()

    if not all( isinstance(schema, SchemaRef) for schema in schema_deps ):
      raise SchemaDeclaredError(
        f"`schema_deps` must be list of SchemaRef: {schema_deps}")

    if not all( isinstance(schema, SchemaRef) for schema in schema_refs ):
      raise SchemaDeclaredError(
        f"`schema_refs` must be list of SchemaRef: {schema_refs}")

    namespace = {
      **namespace,
      '_p_schema_resolved' : False,
      # all dependencies of this schema
      '_p_schema_deps' : schema_deps,
      # schemas referencing this schema as a dependency
      '_p_schema_refs' : schema_refs }

    cls = super().__new__( mcls, name, bases, namespace )

    return cls

  #-----------------------------------------------------------------------------
  def __init__( cls,
    name,
    bases,
    namespace,
    schema_deps = None,
    schema_refs = None ):

    super().__init__( name, bases, namespace )

  #-----------------------------------------------------------------------------
  @property
  def schema_defined( cls ):
    return cls._p_schema_resolved and len(cls._p_schema_deps) == 0

  #-----------------------------------------------------------------------------
  def schema_depends( cls, schema ):
    """Reference another schema having this schema as a dependency

    Parameters
    ----------
    schema : :class:`SchemaDep <partis.schema_meta.schema.SchemaDep`
    """
    if schema not in cls._p_schema_refs:
      cls._p_schema_refs.append( schema )

    if cls.schema_defined:
      schema.schema_resolved( cls )

  #-----------------------------------------------------------------------------
  def _schema_resolved( cls ):
    """Internal method called once all dependencies have been resolved.
    """
    raise NotImplementedError(f"Not implemented by {cls}")

  #-----------------------------------------------------------------------------
  def schema_resolved( cls, schema = None ):
    """Schema previously referenced as a dependency has been resolved

    Parameters
    ----------
    schema : :class:`SchemaRef <partis.schema_meta.schema.SchemaRef`
    """

    if cls._p_schema_resolved:
      return True

    if schema:
      if schema in cls._p_schema_deps:
        cls._p_schema_deps.remove( schema )

      if schema.schema in cls._p_schema_deps:
        cls._p_schema_deps.remove( schema.schema )

    if not cls.schema_resolvable():
      # still at least one un-resolved dependency
      return False

    # fully resolved
    cls._p_schema_resolved = True

    with schema_errors(
      msg = f"Resolution failed: {cls.__module__}.{cls.__name__}",
      loc = cls.loc if hasattr(cls, 'loc') else None,
      cls = SchemaDefinitionError ):

      cls._schema_resolved()

    refs = cls._p_schema_refs
    cls._p_schema_refs = list()

    for schema in refs:
      # signal schemas referencing this schema that it is now defined
      schema.schema_resolved( cls )


    return True

  #-----------------------------------------------------------------------------
  def schema_resolvable( cls, checking = None ):
    """Schema previously referenced as a dependency has been resolved

    Parameters
    ----------
    schema : :class:`SchemaRef <partis.schema_meta.schema.SchemaRef`
    """
    if checking is None:
      checking = list()

    if cls.schema in checking:
      return True

    checking.append(cls.schema)

    for _schema in cls._p_schema_deps:
      if not _schema.schema_resolvable( checking = checking ):
        return False

    return True

  #-----------------------------------------------------------------------------
  def _schema_hash_node( cls ):
    """Internal method to format the hashes of the class state contributing
    to the hash of the schema, **not** including the relative
    'connections' to other schemas.

    .. note::

      Each line may be formatted in any manner that uniquely identifies the
      *relevant* state of the schema as it relates to validating data.

    Returns
    -------
    list[str]

    See Also
    --------
    * :meth:`schema_hash_node`
    """

    return list()

  #-----------------------------------------------------------------------------
  @cached_property
  def schema_hash_node( cls ):
    """urlsafe Base64 encoded hash of this schema, **not** including connections
    (dependencies) to other schemas.

    .. note::

      The hash might not include all state information of the object, such as
      documentation or other data that might be used for informing user
      interfaces.
      The intention is the hash only identifies unique data **structure**.

    Returns
    -------
    str
    """

    return hash_sha256(
      '\n'.join( cls._schema_hash_node() ).encode('utf-8') )[0]

  #-----------------------------------------------------------------------------
  def _schema_hash_children( cls ):
    """Internal method to get the connected 'child' schema classes that
    contribute to the hash of the  schema.

    See Also
    --------
    * :meth:`schema_hash`

    Returns
    -------
    list[ :class:`SchemaRef <partis.schema_meta.schema.SchemaRef` ]
    """
    return list()

  #-----------------------------------------------------------------------------
  def _schema_hash( cls, visited = None ):
    """Internal method to format all connections that contribute to the hash of
    the  schema.

    The algorithm is generalized to handle reference cycles by enumerating all
    recursively referenced schemas in a **depth-first order** with no repeats.
    Each line is formatted as:

      `[node index], [hash of node], [list of child node indices]`

    For example, a hash starting at an arbitrary node marked as ``node(0)`` that
    includes a dependency cycle:

    .. code-block::

          0------<------
          |            |
      --<--->--        |
      |       |        |
      1       3        |
      |       |        |
      2   --<--->--    |
          |       |    |
          4       5-->--

    .. code-block::

      2, hash(node(2)), []
      1, hash(node(1)), [2]
      4, hash(node(4)), []
      5, hash(node(5)), [0]
      3, hash(node(3)), [4, 5]
      0, hash(node(0)), [1, 3]


    Parameters
    ----------
    visited : None | list[ :class:`SchemaRef <partis.schema_meta.schema.SchemaRef` ]
      Internally tracks the schema classes that have already been visited in
      a depth-first search

      .. note::

        This list will be updated by the method call with all nodes visited that
        where not initially in the list.

    Returns
    -------
    list[str]

    See Also
    --------
    * :meth:`schema_hash`

    Note
    ----
    The hash is only of the relative sub-graph starting at the schema the
    method is initially called on, and not necessarily of the entire graph of
    all schemas.

    Note
    ----
    The `schema_hash` attribute should be used to access the generated
    hash.

    :meta public:
    """

    if visited is None:
      visited = list()

    lines = list()

    if cls in visited:
      # base case where a cycle exists
      return lines

    visited.append( cls )

    idx = len(visited) - 1

    edges = list()

    for v in cls._schema_hash_children():

      lines.extend( v.schema._schema_hash( visited = visited ) )

      edges.append( visited.index( v.schema ) )

    # this node defined by its visit index, hash of non-connecting data, and
    # the indices of the nodes of outgoing connections
    lines.append( f'{idx}, {cls.schema_hash_node}, {edges}' )

    return lines

  #-----------------------------------------------------------------------------
  @cached_property
  def schema_hash( cls ):
    """urlsafe Base64 encoded hash of this schema, **including** connections
    to other schemas.

    The ``schema_hash`` is computed to be unique to the combination of all
    'relevent' attributes and relative connections between this schema and all
    other referenced schemas.
    The algorithm is generalized to handle reference cycles by enumerating all
    recursively referenced schemas in a depth-first order.

    Returns
    -------
    str

    """

    return hash_sha256(
      '\n'.join( cls._schema_hash() ).encode('utf-8') )[0]

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaDeclared( SchemaDep ):
  """Base class for all schema forward declarations

  A schema declared acts as a forward declaration of a schema class, where
  a reference to the schema is needed before the schema class has actually been
  defined.
  A declared may only be defined by a *single* eventual schema class.

  """

  #-----------------------------------------------------------------------------
  def __new__( mcls,
    name,
    bases,
    namespace,
    schema_deps = None,
    schema_refs = None ):

    namespace = {
      **namespace,
      '_p_schema': None }

    cls = super().__new__( mcls,
      name,
      bases,
      namespace,
      schema_deps = schema_deps,
      schema_refs = schema_refs )

    return cls

  #-----------------------------------------------------------------------------
  def __init__( cls,
    name,
    bases,
    namespace,
    schema_deps = None,
    schema_refs = None ):

    super().__init__( name, bases, namespace )

  #-----------------------------------------------------------------------------
  @property
  def schema( cls ):

    if not cls._p_schema:
      raise SchemaDeclaredError(
        f"Schema declared but not defined: {cls}")

    return cls._p_schema

  #-----------------------------------------------------------------------------
  @property
  def schema_defined( cls ):
    return cls._p_schema and super().schema_defined and cls._p_schema.schema_defined

  #-----------------------------------------------------------------------------
  def schema_resolvable( cls, checking = None ):
    return (
      cls._p_schema is not None
      and cls._p_schema.schema_resolvable(checking = checking) )

  #-----------------------------------------------------------------------------
  def _schema_resolved( cls ):
    cls.schema._schema_resolved()

  #-----------------------------------------------------------------------------
  def _schema_declared( cls, schema ):
    """Internal method for validating when a schema is to be defined

    Parameters
    ----------

    Returns
    """

    if cls._p_schema:
      raise SchemaDeclaredError(
        f"Schema already defined")

    if not is_schema( schema ):
      raise SchemaDeclaredError(
        f"Schema declared must be defined with a schema class: {fmt_base_or_type(schema)}")

    schema = schema.schema

    if not isinstance(schema, SchemaDep):
      raise SchemaDeclaredError(
        f"Schema declared must be defined with a schema class: {fmt_base_or_type(schema)}")

    return schema

  #-----------------------------------------------------------------------------
  def schema_declared( cls, schema ):
    """Defines the schema for this schema declared.

    .. note::

      This method should return the concrete schema, for example if the provided
      is a reference to the actual schema to be used to define it, since
      the declared must act as the reference to the concrete schema and not another
      reference.

    Parameters
    ----------
    schema : :class:`SchemaRef <partis.schema_meta.schema.SchemaRef`

    Returns
    -------
    :class:`SchemaDep <partis.schema_meta.schema.SchemaDep`
    """

    schema = cls._schema_declared( schema )

    cls._p_schema = schema

    cls._p_schema_deps.append( cls._p_schema )
    cls._p_schema.schema_depends( cls )

  #-----------------------------------------------------------------------------
  def _get_doc( cls,
    noindex = False,
    depth = 0,
    max_depth = 2 ):

    return cls.schema._get_doc(
      noindex = True,
      depth = depth,
      max_depth = max_depth )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class Schema( SchemaDep ):
  """Base class for all schemas

  Parameters
  ----------
  name : str
    Class name of type
  bases : list[type]
    Base classes of type
  namespace : dict[str, object]
  evaluated : None | :class:`Evaluated <partis.schema.eval.Evaluated>`
    Class capable to transforming raw source data (such as Python expressions)
    into the appropriate data type.
    The `val` is checked using `evaluated.check`.
    If it cannot be interpreted as an expression, the value is validated as is.
    If not specified, then source data will not be considered as an expression.
  default_val : None | :class:`OptionalType <partis.utils.special.OptionalType>` | :class:`RequiredType <partis.utils.special.RequiredType>` | :class:`DerivedType <partis.utils.special.DerivedType>` | object
    The default `val` (value) of this schema in the case of missing source data.
    A default of `None` (or `OptionalType`) means the value is optional and will
    be None if source data is missing.
    A default of :class:`RequiredType <partis.utils.special.RequiredType>`
    means the value is required in the source data and
    will raise a :class:`SchemaValidationError <partis.schema_meta.base.SchemaValidationError>`
    if value is missing.
    A default of :class:`DerivedType <partis.utils.special.DerivedType>`
    will attempt to construct a default value from the
    default values of the `struct` items, but will raise a
    :class:`SchemaDefinitionError <partis.schema_meta.base.SchemaDefinitionError>`
    if any items themselves have required values.
    Otherwise, items will be validated against the primitve schema.
  default_eval: :class:`RequiredType <partis.utils.special.RequiredType>` | object
    The default value to use if an evaluated expression results in `None`.
    The purpose of this value is that the `default_val` itself may be an expression
    that could return None, which would reference back to the `default_val` causing
    an un-resolvable loop.
    If that would occur, then this value will be used as the result of the expression
    instead of the `default_val` expression.
    If `default_val` is anything other than an expression, specifying this will raise
    an error to ensure a single source of truth for the resulting value.
  init_val : None | object
    The initial value that is used when creating an editable template for the value.
    This differs from `default_val` in that this value is *not* used to fill in
    any missing source data.
    If `None`, `default_val` will be used, or a value will be derived
    from the schema to produce an initial value that will pass validation.
    However, `init_val` may specified independently from `default_val`.
  preset_vals : None | list[ :class: `PresetValue <partis.schema_meta.base.PresetValue>`]
  schema_deps : None | list[ :class:`SchemaRef <partis.schema_meta.schema.SchemaRef>` ]
    Other schemas which this schema depends upon to be defined.
  schema_refs : None | list[ :class:`SchemaRef <partis.schema_meta.schema.SchemaRef>` ]
    Other schemas which depend on (defined with) this schema.
  valued_type : None | type
    The type used to instantiate values that have been validated ('decoded')
    by this schema.
    This may be initially undefined, by the schema will not be defined/resolved
    until it is set using the ``valued_type`` setter.
  doc : None | str
    Description of this schema
  loc : None | :class:`Loc <partis.schema_meta.base.Loc>`
    Location information of source data (E.G. file, line/column number)

  Raises
  ------
  SchemaDefinitionError
    If the schema definition is not valid

  """


  #-----------------------------------------------------------------------------
  def __new__( mcls,
    name,
    bases,
    namespace,
    evaluated = None,
    default_val = None,
    default_eval = None,
    init_val = None,
    preset_vals = None,
    schema_deps = None,
    schema_refs = None,
    valued_type = None,
    doc = None,
    loc = None ):


    if default_val is None:
      default_val = required

    if any(
      check( default_val )
      for check in [
        is_valued_type,
        is_evaluated ] ):

      default_val = default_val._encode

    if is_special( default_val ):

      if not any(
        check( default_val )
        for check in [
          is_required,
          is_optional,
          is_derived ] ):

        raise SchemaDefinitionError(
          f"`default_val` special values may only be `optional`, `required`, or `derived`: {default_val}",
          loc = loc )

    if evaluated is None:
      from partis.schema.eval import NotEvaluated
      evaluated = NotEvaluated

    if not is_evaluated_class( evaluated ):
      raise SchemaDefinitionError(
        f"`evaluated` must be subclass of `Evaluated`: {fmt_base_or_type(evaluated)}",
        loc = loc )


    if default_eval is None:
      default_eval = required

    if any(
      check( default_eval )
      for check in [
        is_valued_type,
        is_evaluated ] ):

      default_eval = default_eval._encode

    if not is_required( default_eval ):

      if not evaluated.check( default_val ):
        raise SchemaDefinitionError(
          f"`default_eval` not needed if `default_val` is not an expression: {default_eval}, {default_val}",
          loc = loc )

      if is_optional( default_eval ):
        raise SchemaDefinitionError(
          f"`default_eval` may not be an `OptionalType`: {default_eval}",
          loc = loc )

      if evaluated.check( default_eval ):
        raise SchemaDefinitionError(
          f"`default_eval` may not be an expression: {default_eval}",
          loc = loc )

    if any(
      check( init_val )
      for check in [
        is_valued_type,
        is_evaluated ] ):

      init_val = init_val._encode

    if is_special( init_val ):
      raise SchemaDefinitionError(
        f"`init_val` may not be a special value: {init_val}",
        loc = loc )

    if preset_vals is None:
      preset_vals = list()

    if not isinstance(preset_vals, list):
      raise SchemaDefinitionError(
        f"`preset_vals` must be a list: {preset_vals}",
        loc = loc )

    preset_vals = copy(preset_vals)

    for i,v in enumerate(preset_vals):
      if not isinstance(v, PresetValue):
        preset_vals[i] = PresetValue(v)

    if valued_type is not None:
      if not isinstance( valued_type, type ):
        raise SchemaDefinitionError(
          f"`valued_type` must be a type: {valued_type}",
          loc = loc )

    # format docstring
    if doc is None:
      doc = ""
    else:
      doc = inspect.cleandoc( doc )


    if '__doc__' in namespace and namespace['__doc__']:
      _doc = inspect.cleandoc( str(namespace['__doc__']) )

      if doc:
        doc += '\n\n' + _doc
      else:
        doc = _doc

    if loc is None:
      loc = Loc()

    namespace = { **namespace, **dict(
      __doc__ = StringFunction(),
      _p_doc = doc,
      _p_evaluated = evaluated,
      _p_default_val = default_val,
      _p_default_eval = default_eval,
      _p_init_val = init_val,
      _p_preset_vals = preset_vals,
      _p_valued_type = valued_type,
      _p_loc = loc )}

    cls = super().__new__(
      mcls,
      name,
      bases,
      namespace,
      schema_deps = schema_deps,
      schema_refs = schema_refs )

    cls.__doc__._func = cls._get_doc

    if schema_deps:
      for schema in schema_deps:
        if schema:
          if isinstance( schema, SchemaDep ):
            schema.schema_depends( cls )

          elif isinstance( schema, SchemaProxy ):
            schema.schema.schema_depends( cls )

          else:
            raise SchemaDefinitionError(
              f"`schema_deps` must be SchemaDep or SchemaProxy: {schema}",
              loc = loc )

    return cls

  #-----------------------------------------------------------------------------
  def __init__( cls,
    name,
    bases,
    namespace,
    evaluated = None,
    default_val = None,
    default_eval = None,
    init_val = None,
    schema_deps = None,
    schema_refs = None,
    valued_type = None,
    doc = None,
    loc = None ):

    super().__init__( name, bases, namespace )

    cls.schema_resolved()

  #-----------------------------------------------------------------------------
  @property
  def schema_defined( cls ):
    return cls.valued_type and super().schema_defined

  #-----------------------------------------------------------------------------
  @property
  def valued_type( cls ):
    return cls._p_valued_type

  #-----------------------------------------------------------------------------
  @valued_type.setter
  def valued_type( cls, valued_type ):

    if valued_type is None:
      return

    if cls._p_valued_type:
      raise SchemaDefinitionError(
        f"Schema valued type already defined: {cls._p_valued_type}",
        loc = cls.loc )

    if not isinstance( valued_type, type ):
      raise SchemaDefinitionError(
        f"`valued_type` must be a type: {valued_type}",
        loc = cls.loc )

    cls._p_valued_type = valued_type

    cls.schema_resolved()

  #-----------------------------------------------------------------------------
  def schema_resolvable( cls, checking = None ):
    return (
      cls.valued_type is not None
      and super().schema_resolvable(checking = checking) )

  #-----------------------------------------------------------------------------
  @property
  def doc( cls ):
    """str : Base documentation
    """
    return cls._p_doc

  #-----------------------------------------------------------------------------
  def _get_attr_doc_lines( cls,
    noindex = False,
    depth = 0,
    max_depth = 2 ):

    return list()

  #-----------------------------------------------------------------------------
  def _get_doc( cls,
    noindex = False,
    depth = 0,
    max_depth = 2 ):


    lines = [
      cls.doc,
      '', '' ]

    if isinstance( cls, Schema ) and cls.schema_defined and depth < max_depth:
      lines.extend( cls._get_attr_doc_lines(
        noindex = noindex,
        depth = depth,
        max_depth = max_depth ) )

    return "\n".join( lines )

  #-----------------------------------------------------------------------------
  @property
  def loc( cls ):
    return cls._p_loc

  #-----------------------------------------------------------------------------
  @property
  def schema( cls ):

    return cls

  #-----------------------------------------------------------------------------
  @property
  def hints( cls ):
    """list[ :class:`SchemaHint <partis.schema_meta.base.SchemaHint>` ] : Definition
    hints.
    """

    hints = [
      SchemaHint(
        f"doc: {cls.doc}" ) ]

    return hints

  #-----------------------------------------------------------------------------
  # def depends( cls, schema ):
  #
  #   if schema not in cls._p_schema_refs:
  #     cls._p_schema_refs.append( schema )
  #
  #   if cls.schema_defined:
  #     schema.resolve( cls )

  #-----------------------------------------------------------------------------
  def _schema_resolved( cls ):

    # fully resolved
    # test-run the generation of values
    for attr in [
      'default_val',
      'default_eval',
      'init_val']:

      with schema_errors(
        msg = f"{cls.__name__}.{attr} not valid",
        loc = cls.loc,
        cls = SchemaDefinitionError ):

        v = getattr( cls.schema, attr )

        if not (
          is_optional( v )
          or is_required( v ) ):

          cls.schema.decode( v, cls.loc )

    for i, preset in enumerate(cls.preset_vals):
      with schema_errors(
        msg = f"{cls.__name__}.preset_vals[{i}] not valid",
        loc = cls.loc,
        cls = SchemaDefinitionError ):

        cls.schema.decode( preset.val, cls.loc )

  #-----------------------------------------------------------------------------
  @property
  def default_val( cls ):
    """object: default value
    """
    return cls._p_default_val

  #-----------------------------------------------------------------------------
  @property
  def default_eval( cls ):
    """object : default value used when the default value is itself an expression
    that returns nothing.
    """
    return cls._p_default_eval

  #-----------------------------------------------------------------------------
  @property
  def init_val( cls ):
    """object : initial value used, even when there is no default value
    """
    return cls._p_init_val

  #-----------------------------------------------------------------------------
  @property
  def preset_vals( cls ):
    """object : preset values that may be used
    """
    return cls._p_preset_vals

  #-----------------------------------------------------------------------------
  @property
  def evaluated( cls ):
    """:class:`Evaluated <partis.schema.eval.Evaluated>` : Evaluation class
    """
    return cls._p_evaluated



import pytest

from pprint import pprint

import itertools

from partis.utils.special import (
  SpecialType )

from partis.schema import (
  required,
  optional,
  BoolPrim,
  IntPrim,
  FloatPrim,
  StrPrim,
  SeqPrim,
  MapPrim,
  UnionPrim,
  SchemaStruct,
  PJCEvaluated,
  SchemaError,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  any_schema,
  BoolValued,
  IntValued,
  FloatValued,
  StrValued,
  SeqValued,
  MapValued,
  SchemaDefinitionError )

from partis.schema.serialize.yaml import (
  loads,
  dumps )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SIMPLE_PRIMS = [
  BoolPrim,
  IntPrim,
  FloatPrim,
  StrPrim ]

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
max_tests = -1
step_tests = 1

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_defined( val ):
  return not ( val is None or val is required or val is optional )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def get_val( val, default_val ):
  if val is None:
    if is_defined(default_val):
      return default_val

  return val

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def should_pass(
  val,
  is_type,
  is_item = None,
  default_val = optional,
  restricted = None,
  evaluated = None,
  min = None,
  max = None ):

  # check if value is to be evaluated
  _is_eval = evaluated is not None and evaluated.check(val)

  # check if value can be interpreted as given type
  if is_type in [ is_bool, is_numeric, is_string ]:

    _is_type = ( ( val is None
        or is_type(val)
        or _is_eval )
      and ( not is_defined( default_val )
        or is_type(default_val) ) )

  elif is_type is is_sequence:

    _is_type = ( ( val is None
        or ( is_sequence(val)
          and ( len(val) == 0 or all( should_pass( val = v, is_type = is_item, default_val = required ) for v in val ) ) )
        or _is_eval )
      and ( not is_defined( default_val )
        or ( is_sequence(default_val)
          and ( len(default_val) == 0
          or all( should_pass( val = v, is_type = is_item, default_val = required ) for v in default_val ) ) ) ) )

  elif is_type is is_mapping:

    _is_type = ( ( val is None
        or ( is_mapping(val)
          and ( len(val) == 0 or all( should_pass( val = v, is_type = is_item, default_val = required ) for v in val.values() ) ) )
        or _is_eval )
      and ( not is_defined( default_val )
        or ( is_mapping(default_val)
          and ( len(default_val) == 0
            or all( should_pass( val = v, is_type = is_item, default_val = required ) for v in default_val.values() ) ) ) ) )

  else:
    assert False

  # check that value is valid
  val = get_val( val, default_val )

  _is_valid = _is_type and not (
    ( val is None and default_val is required )
    or ( restricted is not None and (
      len(restricted) == 0
      or any( not should_pass( val = v, is_type = is_type, is_item = is_item, default_val = required, min = min, max = max ) for v in restricted )
      or ( is_defined(default_val) and default_val not in restricted )
      or ( not _is_eval and is_defined(val) and val not in restricted ) ) )
    or ( min is not None and (
      ( is_defined(default_val) and default_val < min )
      or ( not _is_eval and is_defined(val) and val < min ) ) )
    or ( max is not None and (
      ( min is not None and max < min )
      or ( is_defined(default_val) and default_val > max )
      or ( not _is_eval and is_defined(val) and val > max ) ) ) )

  return _is_valid

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_default_val():
  class DummyType( SpecialType ):
    pass

  for prim in SIMPLE_PRIMS:
    with pytest.raises( SchemaDefinitionError ):
      s = prim(
        default_val = DummyType() )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_restricted():

  for prim in SIMPLE_PRIMS:
    with pytest.raises( SchemaDefinitionError ):
      s = prim(
        restricted = "abcd" )

    with pytest.raises( SchemaDefinitionError ):
      s = prim(
        restricted = list() )




import os
import subprocess
import shutil
from timeit import default_timer as timer

import logging
log = logging.getLogger(__name__)

from partis.schema import (
  Loc,
  required,
  optional,
  derived,
  StrPrim,
  StrValued )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class ColorValued( StrValued ):
  #-----------------------------------------------------------------------------
  @property
  def rgb(self):
    return int(self[1:3], 16), int(self[3:5], 16), int(self[5:7], 16)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class Color(StrPrim):
  #-----------------------------------------------------------------------------
  def __new__( mcls,
    name,
    bases,
    namespace,
    default_val,
    **kwargs ):

    if default_val is None:
      default_val = '#000000'

    return super().__new__( mcls,
      name = name,
      bases = bases,
      namespace = namespace,
      char_case = 'upper',
      strip = True,
      pattern = r"#[0-9A-F]{6}",
      nonempty = True,
      max_lines = 1,
      max_cols = 7,
      valued_type = ColorValued,
      default_val = default_val,
      **kwargs )




from .eval import (
  ProviderSupport,
  ProviderMeta,
  Provider,
  EvaluatedMeta,
  Evaluated,
  NotEvaluated,
  EvaluatedContext,
  ConstEvaluatedContext )

from .eval_py import (
  PyEvaluated,
  PyEvaluatedRestricted,
  EvalFunc )

from .eval_cheetah import CheetahEvaluated
from .eval_pjc import PJCEvaluated



# -*- coding: UTF-8 -*-

import logging
log = logging.getLogger(__name__)

from copy import copy
import inspect
import re

from partis.utils import (
  odict,
  fmt_base_or_type,
  fmt_iterable_or,
  fmt_attr_doc )

from partis.pyproj import (
  hash_sha256 )

from partis.schema_meta.base import (
  SchemaError,
  SchemaNameError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaHint,
  Bias,
  Loc,
  assert_valid_name,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated,
  is_valued,
  is_optional,
  is_required )

from partis.schema.valued import (
  FloatValued )

from . import (
  SchemaPrimDeclared,
  SchemaPrim )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class FloatPrimDeclared( SchemaPrimDeclared ):
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class FloatPrim( SchemaPrim ):
  """Primitive for floating point values

  Parameters
  ----------
  **kwargs : arguments passed to :class:`SchemaPrim <partis.schema.prim.SchemaPrim>`
  """

  #-----------------------------------------------------------------------------
  def __new__( mcls,
    name,
    bases,
    namespace,
    min = None,
    max = None,
    **kwargs ):

    loc = kwargs.get('loc', None)

    if min is not None:
      min = float(min)

    if max is not None:
      max = float(max)

      if not ( min is None or max >= min ):
        raise SchemaDefinitionError(
          f"`FloatPrim` max must be greater than or equal to min ({min}): {max}",
          loc = loc )

    namespace = { **namespace, **dict(
      _p_min = min,
      _p_max = max )}


    return super().__new__( mcls,
      name = name,
      bases = bases,
      namespace = namespace,
      valued_type = FloatValued,
      **kwargs )

  #-----------------------------------------------------------------------------
  def __init__( cls,
    name,
    bases,
    namespace,
    min = None,
    max = None,
    **kwargs ):

    super().__init__(
      name = name,
      bases = bases,
      namespace = namespace,
      **kwargs )

  #-----------------------------------------------------------------------------
  def _get_attr_doc_lines( cls,
    noindex = False,
    depth = 0,
    max_depth = 2 ):

    lines = super()._get_attr_doc_lines(
      noindex = noindex,
      depth = depth,
      max_depth = max_depth )

    lines.extend([
      fmt_attr_doc(
        name = attr,
        typename = type( getattr( cls, attr ) ),
        val = getattr( cls, attr ),
        prefix = 'schema',
        noindex = noindex )
      for attr in [
        'min',
        'max' ] ])

    return lines

  #-----------------------------------------------------------------------------
  def _schema_hash_node( cls ):
    lines = super()._schema_hash_node()

    for attr in [
      'min',
      'max' ]:

      val = getattr( cls, attr )
      hash = None

      if hasattr( val, 'schema_hash' ):
        hash = val.schema_hash
      else:
        hash = hash_sha256( str(val).encode('utf-8') )[0]

      lines.append( f'{attr}, {hash}' )

    return lines

  #-----------------------------------------------------------------------------
  @property
  def hints( cls ):
    hints = super().hints

    if cls.min is not None:
      hints.append( SchemaHint(
        "min: float",
        hints = f"value >= {cls.min}" ) )

    if cls.max is not None:
      hints.append( SchemaHint(
        "max: float",
        hints = f"value <= {cls.max}" ) )

    return hints

  #-----------------------------------------------------------------------------
  @property
  def min( cls ):
    return cls._p_min

  #-----------------------------------------------------------------------------
  @property
  def max( cls ):
    return cls._p_max

  #-----------------------------------------------------------------------------
  def validate( cls,
    val,
    loc = None,
    bias = None ):

    bias = Bias(bias)

    if not is_numeric( val ):
      raise SchemaValidationError(
        f"Must be a `numeric`: {fmt_base_or_type(val)}",
        loc = loc )

    bias += 1
    _val = float( val.real )

    if _val != val:
      raise SchemaValidationError(
        f"Casting to float does not preserve value: {fmt_base_or_type(val)}",
        loc = loc )

    val = _val

    if cls.min is not None:
      bias += 1

      if val < cls.min:
        raise SchemaValidationError(
          f"Must be >= {cls.min}: {fmt_base_or_type(val)}",
          loc = loc )

    if cls.max is not None:
      bias += 1

      if val > cls.max:
        raise SchemaValidationError(
          f"Must be <= {cls.max}: {fmt_base_or_type(val)}",
          loc = loc )

    return super().validate(
      val = val,
      loc = loc,
      bias = bias )

  #-----------------------------------------------------------------------------
  @property
  def init_val( cls ):
    val = super().init_val

    if cls.min is not None and val < cls.min:
      val = cls.min

    if cls.max is not None and val > cls.max:
      val = cls.max

    return val



# -*- coding: UTF-8 -*-
import os
import sys
import inspect
import weakref
from inspect import getframeinfo, stack
import traceback
import linecache
from copy import copy, deepcopy
import importlib

import logging
log = logging.getLogger(__name__)

from Cheetah.Template import Template

# linting cheetah source
from Cheetah import (
  NameMapper,
  Parser )

from partis.utils import (
  indent_lines,
  line_segment )

from partis.schema_meta.base import (
  SchemaHint,
  Loc,
  SchemaEvaluationError )

from .eval import (
  ProviderSupport,
  Provider,
  Evaluated )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class CheetahProvider( Provider ):
  """Evaluates cheetah template expressions
  """

  TAG_TMPL_CHEETAH = "$tmpl:cheetah"

  #-----------------------------------------------------------------------------
  def __init__( self ):
    super().__init__()

    self._supported = dict(
      tmpl_cheetah = ProviderSupport(
        name = "Cheetah template",
        lexer = 'cheetah',
        doc = f"escaped string: `{self.TAG_TMPL_CHEETAH} ...`") )

  #-----------------------------------------------------------------------------
  @property
  def supported( self ):
    return self._supported

  #-----------------------------------------------------------------------------
  def check( self, src ):

    if isinstance( src, str ):

      if src.startswith("$tmpl:cheetah"):
        # cheetah template
        tmpl = src[len("$tmpl:cheetah"):]

        if tmpl.startswith("\n"):
          tmpl = tmpl[1:]

        return self.supported['tmpl_cheetah'], tmpl

    return None

  #-----------------------------------------------------------------------------
  def escaped( self, support, src ):
    if support is self.supported['tmpl_cheetah']:
      # cheetah template
      return f"{self.TAG_TMPL_CHEETAH}\n{src}"


    raise ValueError(f"`support` must be one of {self.supported.values()}: {support}")

  #-----------------------------------------------------------------------------
  def eval( self,
    schema,
    src,
    loc = None,
    locals = None,
    module = None,
    logger = None ):

    if locals is None:
      locals = dict()

    if logger is None:
      logger = log

    supported = self.check( src )

    if not supported:
      raise SchemaEvaluationError(f"No evaluation support: {src}")

    support, src = supported

    # ensure currend directory is preserved after executing script
    cwd = os.path.abspath( os.getcwd() )

    try:
      if support is self.supported['tmpl_cheetah']:
        # cheetah template

        return str( Template( src, searchList = [ locals ] ) )

      assert False, str(support)

    finally:
      os.chdir( cwd )

  #-----------------------------------------------------------------------------
  def lint( self,
    schema,
    src,
    loc = None,
    locals = None,
    module = None,
    logger = None ):

    supported = self.check( src )

    if not supported:
      raise SchemaEvaluationError(f"No evaluation support: {src}")

    support, src = supported

    if support is self.supported['tmpl_cheetah']:
      # cheetah template

      return self.lint_cheetah( src, loc, locals )

  #-----------------------------------------------------------------------------
  def lint_cheetah( self, src, loc, locals ):
    if loc is None:
      loc = Loc()
      
    hints = list()

    try:

      val = str( Template( src, searchList = [ locals ] ) )

    except NameMapper.NotFound as e:

      hints.append( SchemaHint(
        str(e),
        loc = loc,
        level = 'warning' ) )

    except Parser.ParseError as e:
      if e.lineno:
        lineno = e.lineno
        col = e.col or 1
        line = e.stream.splitlines()[lineno-1]

      else:
        lineno, col, line = e.stream.getRowColLine()


      msg = e.msg

      loc = loc.replace( line = lineno, col = col )

      offset = col - 1
      idx = lineno - 1

      if line:
        msg += f"\n{indent_lines(2, line)}\n"
        msg += '  ' + ' '*offset + '^'

      hints.append( SchemaHint(
        msg,
        loc = loc,
        level = 'error' ) )


    return hints

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
cheetah_provider = CheetahProvider()

class CheetahEvaluated( Evaluated, provider = cheetah_provider ):
  pass



# -*- coding: UTF-8 -*-

import logging
log = logging.getLogger(__name__)

import io
from abc import ABCMeta
from copy import copy
import inspect
import re

from partis.utils import (
  odict,
  fmt_base_or_type,
  fmt_iterable_or,
  StringFunction,
  indent_lines,
  _fmt_class_name,
  fmt_class_name,
  fmt_attr_doc )

from partis.pyproj import (
  hash_sha256 )

from partis.utils.special import (
  required,
  optional )

from partis.schema_meta.schema import (
  Schema,
  SchemaDeclared )

from partis.schema_meta.prim import (
  SchemaPrimMeta )

from partis.schema_meta.base import (
  SchemaError,
  SchemaNameError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaHint,
  Bias,
  Loc,
  assert_valid_name,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated,
  is_evaluated_class,
  is_valued,
  is_valued_type,
  is_special,
  is_optional,
  is_required,
  any_schema )

from partis.schema.eval import (
  NotEvaluated )

from partis.schema.valued import (
  Valued )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaPrimDeclared( SchemaDeclared, metaclass = SchemaPrimMeta ):
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaPrim( Schema, metaclass = SchemaPrimMeta ):
  """Base class for all schema primitives

  Parameters
  ----------
  name : str
    Class name of type
  bases : list[type]
    Base classes of type
  namespace : dict[str, object]
  restricted : None | list[object]
    Restricts the possible decoded values to those given in the list, in addition to being
    the correct type.
    If given, items will be validated against the primitve schema.
  evaluated : None | :class:`Evaluated <partis.schema.eval.Evaluated>`
    Class capable to transforming raw source data (such as Python expressions)
    into the appropriate data type.
    The `val` is checked using `evaluated.check`.
    If it cannot be interpreted as an expression, the value is validated as is.
    If not specified, then source data will not be considered as an expression.
  default_val : None | :class:`OptionalType <partis.utils.special.OptionalType>` | :class:`RequiredType <partis.utils.special.RequiredType>` | :class:`DerivedType <partis.utils.special.DerivedType>` | object
    The default `val` (value) of this schema in the case of missing source data.
    A default of `None` (or `OptionalType`) means the value is optional and will
    be None if source data is missing.
    A default of :class:`RequiredType <partis.utils.special.RequiredType>`
    means the value is required in the source data and
    will raise a :class:`SchemaValidationError <partis.schema_meta.base.SchemaValidationError>`
    if value is missing.
    A default of :class:`DerivedType <partis.utils.special.DerivedType>`
    will attempt to construct a default value from the
    default values of the `struct` items, but will raise a
    :class:`SchemaDefinitionError <partis.schema_meta.base.SchemaDefinitionError>`
    if any items themselves have required values.
    Otherwise, items will be validated against the primitve schema.
  default_eval: :class:`RequiredType <partis.utils.special.RequiredType>` | object
    The default value to use if an evaluated expression results in `None`.
    The purpose of this value is that the `default_val` itself may be an expression
    that could return None, which would reference back to the `default_val` causing
    an un-resolvable loop.
    If that would occur, then this value will be used as the result of the expression
    instead of the `default_val` expression.
    If `default_val` is anything other than an expression, specifying this will raise
    an error to ensure a single source of truth for the resulting value.
  init_val : None | object
    The initial value that is used when creating an editable template for the value.
    This differs from `default_val` in that this value is *not* used to fill in
    any missing source data.
    If `None`, `default_val` will be used, or a value will be derived
    from the schema to produce an initial value that will pass validation.
    However, `init_val` may specified independently from `default_val`.
  preset_vals : None | list[ :class: `PresetValue <partis.schema_meta.base.PresetValue>`]
  doc : None | str
    Description or interpretation of the decoded value.
  """

  #-----------------------------------------------------------------------------
  def __new__( mcls,
    name,
    bases,
    namespace,
    declared = None,
    restricted = None,
    evaluated = None,
    default_val = required,
    default_eval = required,
    init_val = None,
    preset_vals = None,
    schema_deps = None,
    valued_type = None,
    loc = None ):

    if declared is not None:
      # defines the given schema declared

      if not isinstance( declared, SchemaPrimDeclared ):
        raise SchemaDefinitionError(
          f"`declared` must be subclass of SchemaPrimDeclared: {fmt_base_or_type(declared)}",
          loc = loc )

    if restricted is not None:
      if not is_sequence( restricted ):
        raise SchemaDefinitionError(
          f"`restricted` must be None or a list: {fmt_base_or_type(restricted)}",
          loc = loc )

      if len(restricted) == 0:
        raise SchemaDefinitionError(
          f"`restricted` must be None or have at least one value",
          loc = loc )

      for i, val in enumerate(restricted):
        if is_valued_type( val ) or is_evaluated( val ):
          restricted[i] = val._encode

    namespace = dict(**namespace,
      _p_restricted = restricted,
      _p_declared = declared )

    cls = super().__new__(
      mcls,
      name,
      bases,
      namespace,
      evaluated = evaluated,
      default_val = default_val,
      default_eval = default_eval,
      init_val = init_val,
      preset_vals = preset_vals,
      schema_deps = schema_deps,
      valued_type = valued_type,
      loc = loc )

    return cls

  #-----------------------------------------------------------------------------
  def __init__( cls,
    name,
    bases,
    namespace,
    declared = None,
    restricted = None,
    evaluated = None,
    default_val = required,
    default_eval = required,
    init_val = None,
    preset_vals = None,
    schema_deps = None,
    valued_type = None,
    loc = None,
    doc = None ):

    # must initialize class before additional validation since it uses the derived
    # class's methods to perform the validation
    super().__init__( name, bases, namespace )

    if cls._p_declared is not None:
      cls._p_declared.schema_declared( cls )

  #-----------------------------------------------------------------------------
  @property
  def default_val( cls ):
    """ :class:`OptionalType <partis.utils.special.OptionalType>` | :class:`RequiredType <partis.utils.special.RequiredType>` | object : The
    default decoded value of this schema in the case of missing source data.
    A default of `None` means the value is optional, and will not have a value
    if source data is missing.
    A default of `required` means the value is required in the source data
    """
    val = cls._p_default_val

    if not (
      is_optional( val )
      or is_required( val ) ):

      val = cls.encode( val, cls.loc )

    return val

  #-----------------------------------------------------------------------------
  @property
  def default_eval( cls ):
    """:class:`RequiredType <partis.utils.special.RequiredType>` | object : The
    default value to use if an evaluated expression results in `None`.
    """
    val = cls._p_default_eval

    if not is_required( val ):

      val = cls.encode( val, cls.loc )

    return val

  #-----------------------------------------------------------------------------
  @property
  def init_val( cls ):
    """object : An initial value to use to create a template value that will
    pass validation
    """
    val = cls._p_init_val

    if val is not None:
      return cls.encode( cls._p_init_val, cls.loc )

    # default_eval checked first, only not-required if `default_val` is an expression

    val = cls.default_eval

    if not is_required( val ):
      return cls.encode( val, cls.loc )

    val = cls.default_val

    if not (
      is_optional( val )
      or is_required( val ) ):

      # only use `default_val` if an actual value, and not an expression

      return cls.encode( val, cls.loc )

    if cls.restricted is not None:
      return cls.encode( cls.restricted[0], cls.loc )

    # use whatever value would be created by the valued object
    if issubclass( cls.valued_type, Valued ):
      return cls.valued_type()._encode

    return None

  #-----------------------------------------------------------------------------
  @property
  def evaluated( cls ):
    """None | :class:`Evaluated <partis.schema.eval.Evaluated>` : Class capable
    to transforming raw source data (such as Python expressions)
    into the appropriate data type.
    """
    return cls._p_evaluated

  #-----------------------------------------------------------------------------
  @property
  def restricted( cls ):
    """None | list[object] : The possible decoded values to those given in the list,
    in addition to being the correct type.
    """
    return cls._p_restricted

  #-----------------------------------------------------------------------------
  def validate( cls,
    val,
    loc = None,
    bias = None ):
    """Validates value against schema definition

    Parameters
    ----------
    val : object
    loc : None | :class:`Loc <partis.schema_meta.base.Loc>`
    bias : None | Bias

    Returns
    -------
    val : object
    bias : Bias

    Raises
    ------
    SchemaValidationError
      If the value is not valid
    """

    bias = Bias(bias)

    if cls.restricted is not None:
      try:
        bias += len(val)
      except TypeError:
        bias += 1

      if val not in cls.restricted:
        raise SchemaValidationError(
          f"Must be any of {fmt_iterable_or(cls.restricted)}: {fmt_base_or_type(val)}",
          loc = loc )

    return val, bias

  #-----------------------------------------------------------------------------
  def _encode( cls,
    val = None,
    loc = None,
    no_defaults = None ):

    if is_valued_type( val ) or is_evaluated( val ):
      if loc is None:
        loc = val._loc

      # NOTE: must still validate value since there are no gaurantees about
      # the original schema
      if not ( val._schema is cls and is_valued(val) ):
        val = val._encode

    if loc is None:
      loc = Loc(path = [cls.__name__])

    if val is None:
      val = cls.default_val

      if is_required( val ):
        raise SchemaValidationError(
          f"Value is required",
          loc = loc,
          hints = cls.doc )

      else:
        return None, loc, False

    # elif no_defaults and val == cls.default_val:
    #   return None, loc, False

    if cls.evaluated.check(val):
      # the value would still need to be evaluated, so cannot further validate
      return val, loc, False

    return val, loc, True

  #-----------------------------------------------------------------------------
  def _decode( cls,
    val = None,
    loc = None,
    bias = None ):
    """Validates value against schema definition

    Parameters
    ----------
    val : object
    loc : None | :class:`Loc <partis.schema_meta.base.Loc>`
    bias : None | Bias

    Returns
    -------
    val : object
    loc : :class:`Loc <partis.schema_meta.base.Loc>`
    validate : bool
    bias : Bias

    Raises
    ------
    SchemaValidationError
      If the value is not valid
    """

    bias = Bias(bias)

    if is_valued_type( val ) or is_evaluated( val ):
      if loc is None:
        loc = val._loc

      val = val._encode

    if loc is None:
      loc = Loc(path = [cls.__name__])

    if val is None:
      bias = Bias(const = True)
      val = cls.default_val

      if is_required( val ):
        raise SchemaValidationError(
          f"Value is required",
          loc = loc,
          hints = cls.doc )

      elif is_optional( val ):
        return None, loc, False, bias

    if cls.evaluated.check(val):
      # value will need to be evaluated, so cannot further validate
      return (
        cls.valued_type(
          val = cls.evaluated(
            schema = cls,
            src = val,
            loc = loc ),
          bias = bias ),
        loc,
        False,
        bias )

    return val, loc, True, bias

  #-----------------------------------------------------------------------------
  def encode( cls,
    val = None,
    loc = None,
    no_defaults = None ):
    """Encodes value to basic Python types, performing validation

    Parameters
    ----------
    val : None | object
      Source data used to initialize values
    loc : None | :class:`Loc <partis.schema_meta.base.Loc>`
      Location information of source data (E.G. file, line/column number)
    no_defaults : bool
      Values that are equivalent to the 'default_val' are removed.

    Returns
    -------
    val : object

    Raises
    ------
    SchemaValidationError
      If the value is not valid
    """

    val, loc, validate = cls._encode(
      val = val,
      loc = loc,
      no_defaults = no_defaults )

    if not validate:
      return val

    return cls.validate(
      val = val,
      loc = loc )[0]

  #-----------------------------------------------------------------------------
  def decode( cls,
    val = None,
    loc = None,
    bias = None ):
    """Decodes value to schema values, performing validation

    Parameters
    ----------
    val : None | object
      Source data used to initialize values
    loc : None | :class:`Loc <partis.schema_meta.base.Loc>`
      Location information of source data (E.G. file, line/column number)
    bias : None | Bias

    Returns
    -------
    val : :class:`Valued <partis.schema.valued.Valued>`

    Raises
    ------
    SchemaValidationError
      If the value is not valid
    """

    bias = Bias(bias)

    val, loc, validate, bias = cls._decode(
      val = val,
      loc = loc,
      bias = bias )

    if not validate:
      return val

    val, bias = cls.validate(
      val = val,
      loc = loc,
      bias = bias )

    return cls.valued_type(
      val = val,
      schema = cls,
      loc = loc,
      bias = bias )

  #-----------------------------------------------------------------------------
  def _get_doc( cls,
    noindex = False,
    depth = 0,
    max_depth = 2 ):

    lines = [
      cls.doc,
      '', '' ]

    if isinstance( cls, SchemaPrim ) and depth < max_depth:
      lines.extend( cls._get_attr_doc_lines(
        noindex = noindex,
        depth = depth,
        max_depth = max_depth ) )

    return "\n".join( lines )

  #-----------------------------------------------------------------------------
  def _get_attr_doc_lines( cls,
    noindex = False,
    depth = 0,
    max_depth = 2 ):

    lines = super()._get_attr_doc_lines(
      noindex = noindex,
      depth = depth,
      max_depth = max_depth )

    lines.extend([
      fmt_attr_doc(
        name = attr,
        typename = type( getattr( cls, attr ) ),
        val = getattr( cls, attr ),
        prefix = 'schema',
        noindex = noindex )
      for attr in [
        'default_val',
        'default_eval',
        'init_val',
        'restricted' ] ])

    lines.append( fmt_attr_doc(
      name = 'evaluated',
      typename = cls.evaluated,
      prefix = 'schema',
      noindex = noindex ) )

    return lines

  #-----------------------------------------------------------------------------
  def _schema_hash_node( cls ):
    lines = super()._schema_hash_node()

    for attr in [
      # TODO: should these be included in the hash? Default values used while
      # decoding seem like non-essential information for validating what was
      # originally encoded.
      # 'default_val',
      # 'default_eval',
      # 'init_val',
      'restricted',
      'evaluated' ]:

      val = getattr( cls, attr )
      hash = None

      if hasattr( val, 'schema_hash_node' ):
        hash = val.schema_hash_node
      else:
        hash = hash_sha256( str(val).encode('utf-8') )[0]

      lines.append( f'{attr}, {hash}' )

    return lines

  #-----------------------------------------------------------------------------
  def _schema_resolved( cls ):

    super()._schema_resolved()

    if cls.restricted:
      for i, val in enumerate( cls.restricted ):
        try:

          cls.decode( val )

        except Exception as e:
          raise SchemaDefinitionError(
            f"Schema `restricted[{i}]` not valid: {val}",
            loc = cls.loc,
            hints = SchemaHint.cast( e ) ) from e


  #-----------------------------------------------------------------------------
  def __call__( cls,
    *args,
    **kwargs ):

    if len(args) == 0:
      val = None
      loc = None
    else:
      if len(args) == 1:
        val = args[0]
        loc = None
      elif len(args) == 2:
        val, loc = args
      else:
        raise ValueError(
          f"positional arguments must be at most `(val, loc)`."
          " All keyword arguments are interpreted as items of `val`")

    if kwargs:
      if val:
        val = {**val, **kwargs}
      else:
        val = kwargs

    return cls.decode(
      val = val,
      loc = loc )



# -*- coding: utf-8 -*-

from partis.utils.sphinx import basic_conf

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# configuration
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

globals().update( basic_conf(
  package = 'partis-schema',
  copyright_year = '2022' ) )

# pylint: disable-next=E0602
intersphinx_mapping['partis.pyproj'] = (
  "https://nanohmics.bitbucket.io/doc/partis/pyproj",
  None )

# pylint: disable-next=E0602
intersphinx_mapping['partis.utils'] = (
  "https://nanohmics.bitbucket.io/doc/partis/utils",
  None )



import logging
log = logging.getLogger( __name__ )

from copy import copy

from partis.utils import (
  ModelHint,
  adict,
  adict_struct,
  odict,
  attrs_modify )


from partis.utils.special import (
  required,
  optional )

from partis.schema_meta.valued import (
  ValuedMeta )

from partis.schema_meta.base import (
  SchemaError,
  SchemaHint,
  Bias,
  Loc,
  assert_valid_name,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_special,
  is_optional,
  is_required,
  is_derived,
  is_similar_value_type,
  is_schema_prim,
  is_schema_declared,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated_class,
  is_evaluated,
  is_valued,
  is_valued_type,
  any_schema )

from partis.schema import (
  EvaluatedContext )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def get_schema_loc( val, schema = None, loc = None ):

  if is_valued_type( val ):
    # copy source information to prevent chaining
    if schema is None:
      schema = val._schema

    if loc is None:
      loc = val._loc

    # NOTE: this is necessary before checking for is_evaluated
    val = val._src

  if is_evaluated( val ):
    if schema is None:
      schema = val._schema

    if loc is None:
      loc = val._loc

  if loc is None:
    loc = Loc()

  return schema, loc

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def get_init_val( val, schema, default_val ):

  if val is None or not (
    is_valued( val )
    or is_sequence( val )
    or is_mapping( val ) ):

    # dummy value
    if schema is None:
      schema, loc = get_schema_loc(
        val = val )

    if not (
      schema is None
      or schema.init_val is None ):

      val = schema.init_val

    else:
      val = default_val

  return val

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def get_src_val( val ):

  if is_valued_type( val ):
    val = val._src

  return val

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class Valued( metaclass = ValuedMeta ):
  """Base class of schema primitive values

  Parameters
  ----------
  val : object
    Source data to be evaluated
  schema : :class:`Schema <partis.schema.struct.SchemaStruct>` | :class:`SchemaPrim <partis.schema.prim.base.SchemaPrim>`
  loc : :class:`Loc <partis.schema_meta.base.Loc>`
    Location information of source data (E.G. file, line/column number)
  """
  #-----------------------------------------------------------------------------
  def __init__( self,
    val = None,
    schema = None,
    loc = None,
    bias = None ):

    if bias is None:
      bias = Bias()

    self._p_bias = bias

    self._p_src = get_src_val( val )

    if not is_valued( self._p_src ):
      self._p_valued = False

    else:
      self._p_valued = True

    if not ( schema is None or is_schema( schema ) ):
      raise SchemaError(
        f"`schema` must be instance of `Schema` or `SchemaPrim`: {type(schema)}")

    if not ( loc is None or isinstance( loc, Loc ) ):
      raise SchemaError(
        f"`loc` must be instance of `Loc`: {type(loc)}")

    schema, loc = get_schema_loc(
      val = val,
      schema = schema,
      loc = loc )

    self._p_schema = schema
    self._p_loc = loc

  #-----------------------------------------------------------------------------
  def model_hint( self ):
    name = self._schema.__name__

    if name.endswith('_schema'):
      name = name[:-7]

    hints = list()

    if is_sequence(self):
      for i,v in enumerate(self):
        hints.append(ModelHint.cast(v))

      return ModelHint(
        f"<{name}>",
        level = 'info',
        hints = hints )

    elif is_mapping(self):
      from partis.schema.prim import (
        SeqPrimDeclared,
        SeqPrim,
        MapPrimDeclared,
        MapPrim )

      for k,v in self.items():
        if v is None:
          continue

        if not is_valued_type(v):
          hints.append(ModelHint(
            f"{k}",
            data = v ))

        elif (
          not is_schema_prim( v._schema )
          or isinstance(v._schema, (
            SeqPrim,
            SeqPrimDeclared,
            MapPrim,
            MapPrimDeclared)) ):


          # Cast 'non-trivial' values to full hint object
          hint = ModelHint.cast(v)
          hint.msg = f"{k}"
          hints.append(hint)

        else:
          # 'trivial' values formatted directly into key's hint
          hints.append(ModelHint(
            f"{k}",
            data = v._encode,
            loc = v._loc ))

      return ModelHint(
        data = f"<{name}>",
        level = 'info',
        loc = self._loc,
        hints = hints )

    else:
      return ModelHint(
        data = self._encode,
        loc = self._loc)

  #-----------------------------------------------------------------------------
  def __rich__( self ):
    return ModelHint.cast(self).fmt( with_rich = True )

  #-----------------------------------------------------------------------------
  def __copy__( self ):
    return type(self)(
      val = self,
      loc = self._loc )

  #-----------------------------------------------------------------------------
  def __eq__( self, other ):

    va = is_evaluated(self)
    vb = is_evaluated(other)

    if va:
      if vb:
        return self._src == other._src
      else:
        return self._src == other

    elif vb:
      return self == other._src

    # neither are evaluated

    return super().__eq__( other )

  #-----------------------------------------------------------------------------
  def __ne__( self, other ):
    return not (self == other)

  #-----------------------------------------------------------------------------
  def __hash__( self ):
    if not is_valued(self):
      return hash(self._src)

    return super().__hash__()

  #-----------------------------------------------------------------------------
  @property
  def _schema( self ):
    return self._p_schema

  #-----------------------------------------------------------------------------
  @property
  def _loc( self ):
    return self._p_loc

  #-----------------------------------------------------------------------------
  @property
  def _valued( self ):
    return self._p_valued

  #-----------------------------------------------------------------------------
  @property
  def _src( self ):
    return self._p_src


  #-----------------------------------------------------------------------------
  @property
  def _bias( self ):
    r"""A metric of how biased (selective) the value is to the ``_schema``

    This may be interpreted as the number of needed independent binary decisions
    (questions, or conditional statements) to determine that the value satisfied
    the schema.

    Returns
    -------
    float
    """
    return self._p_bias

  #-----------------------------------------------------------------------------
  @property
  def _encode( self ):
    if self._valued:
      return self._src

    if is_evaluated( self._src ):
      return self._src._encode

    assert False


  #-----------------------------------------------------------------------------
  def _lint( self,
    context = None,
    logger = None ):
    """Lints source

    Parameters
    ----------
    context : None | :class:`EvaluatedContext` | list[ :class:`EvaluatedContext` ]
      Sets the context for which this expression is to be evaluated.
      The expression is evaluated only when the specified context is equivalent
      to the one for this evaluated class.
    logger : logging.Logger
      A logger object to direct expression logs/prints instead of stdout

    Returns
    -------
    : list[SchemaHint]

    See Also
    --------
    :class:`Evaluated <partis.schema.eval.Evaluated>`
    """

    hints = list()

    if self._valued:
      return hints

    if is_evaluated( self._p_src ):

      hints.extend( self._p_src._lint(
        context = context,
        logger = logger ) )

    elif is_sequence( self._p_src ):

      # local variables for items in list or dictionary
      # NOTE: parent of items is the current value, parent of current will be
      # parent of the item parent

      for i, v in enumerate(self._p_src):
        if not is_valued(v):

          if isinstance( context, EvaluatedContext ):
            _context = context(
              schema = self._schema,
              parent = self,
              key = i )

          else:
            _context = context

          hints.extend( v._lint(
            context = _context,
            logger = logger ) )

    elif is_mapping( self._p_src ):
      # local variables for items in list or dictionary
      # NOTE: parent of items is the current value, parent of current will be
      # parent of the item parent

      for k, v in self._p_src.items():

        if not is_valued(v):

          if isinstance( context, EvaluatedContext ):
            _context = context(
              schema = self._schema,
              parent = self,
              key = k )

          else:
            _context = context

          hints.extend( v._lint(
            context = _context,
            logger = logger ) )

    return hints


  #-----------------------------------------------------------------------------
  def _eval( self,
    context = None,
    logger = None ):
    """Evaluates source

    Parameters
    ----------
    context : None | :class:`EvaluatedContext` | list[ :class:`EvaluatedContext` ]
      Sets the context for which this expression is to be evaluated.
      The expression is evaluated only when the specified context is equivalent
      to the one for this evaluated class.
    logger : logging.Logger
      A logger object to direct expression logs/prints instead of stdout

    Returns
    -------
    val : :class:`Valued <partis.schema.valued.Valued>`
      Decoded value resulting from evaluation.
      If there are no un-evaluated values, then will return a shallow copy of
      the current data.

    See Also
    --------
    :class:`Evaluated <partis.schema.eval.Evaluated>`
    """


    if self._valued:
      return copy(self)

    if is_evaluated( self._p_src ):

      val = self._p_src._eval(
        context = context,
        logger = logger )

    elif is_sequence( self._p_src ):

      # local variables for items in list or dictionary
      # NOTE: parent of items is the current value, parent of current will be
      # parent of the item parent
      val = list()

      for i, v in enumerate(self._p_src):
        if is_valued(v):
          val.append( v )

        else:
          if isinstance( context, EvaluatedContext ):
            _context = context(
              schema = self._schema,
              parent = val,
              key = i )

          else:
            _context = context

          val.append( v._eval(
            context = _context,
            logger = logger ) )

    elif is_mapping( self._p_src ):
      # local variables for items in list or dictionary
      # NOTE: parent of items is the current value, parent of current will be
      # parent of the item parent
      val = adict()

      for k, v in self._p_src.items():

        if is_valued(v):
          val[k] = v

        else:
          if isinstance( context, EvaluatedContext ):
            _context = context(
              schema = self._schema,
              parent = val,
              key = k )

          else:
            _context = context

          val[k] = v._eval(
            context = _context,
            logger = logger )


    if is_optional(val):
      return val

    if is_valued_type( val ):
      val._p_loc = self._loc

      return val

    return type(self)(
      val = val,
      schema = self._schema,
      loc = self._loc )




import pytest

from partis.schema import (
  SchemaError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaValidationError,
  SchemaEvaluationError,
  SchemaHint,
  Loc,
  required,
  optional,
  derived,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_optional,
  is_required,
  is_derived,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated_class,
  is_evaluated,
  is_valued,
  is_valued_type,
  any_schema,
  BoolValued,
  IntValued,
  FloatValued,
  StrValued,
  SeqValued,
  MapValued,
  PyEvaluated,
  PJCEvaluated,
  PassPrim,
  BoolPrim,
  IntPrim,
  FloatPrim,
  StrPrim,
  SeqPrim,
  UnionPrim,
  MapPrim,
  StructValued )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class NoRaise:
  def __enter__(self):
    return self

  def __exit__(self, type, value, traceback):
    pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NoError = None

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def _cases( ):


  # test using metaclass arguments
  class MyType1( StructValued ):
    schema = dict(
      tag = 'my_type',
      struct = [
        ( 'a', BoolPrim() ),
        ( 'b', IntPrim() ),
        ( 'c', FloatPrim() ),
        ( 'd', StrPrim() ),
        ( 'e', SeqPrim(
          item = IntPrim() ) ),
        ( 'f', MapPrim(
          item = StrPrim() ) )  ],
      evaluated = PJCEvaluated )


  # test using class namespace arguments
  class MyType2( StructValued ):
    schema = dict(
      tag = 'my_type',
      evaluated = PJCEvaluated )

    a = BoolPrim()
    b = IntPrim()
    c = FloatPrim()
    d = StrPrim()
    e =  SeqPrim(
      item = IntPrim() )

    f = MapPrim(
      item = StrPrim() )

  base_cases = [
    ( NoError,
      BoolPrim(
        evaluated = PJCEvaluated ),
      BoolValued,
      "1 == 1",
      True ),
    ( SchemaEvaluationError,
      BoolPrim(
        evaluated = PJCEvaluated ),
      BoolValued,
      "1",
      True ),
    ( NoError,
      IntPrim(
        evaluated = PJCEvaluated ),
      IntValued,
      "12 + 30",
      42 ),
    ( SchemaEvaluationError,
      IntPrim(
        evaluated = PJCEvaluated ),
      IntValued,
      "3/2",
      1 ),
    ( SchemaEvaluationError,
      IntPrim(
        evaluated = PJCEvaluated ),
      IntValued,
      "'42",
      42 ),
    ( NoError,
      FloatPrim(
        evaluated = PJCEvaluated ),
      FloatValued,
      "2.0 * 4.0",
      8.0 ),
    ( NoError,
      FloatPrim(
        evaluated = PJCEvaluated ),
      FloatValued,
      "2 * 4",
      8 ),
    ( SchemaEvaluationError,
      FloatPrim(
        evaluated = PJCEvaluated ),
      FloatValued,
      "'8.0'",
      8.0 ),
    ( NoError,
      StrPrim(
        evaluated = PJCEvaluated ),
      StrValued,
      "\"hello world\"",
      "hello world" ),
    ( SchemaEvaluationError,
      StrPrim(
        evaluated = PJCEvaluated ),
      StrValued,
      "8.0",
      "8.0" ),
    ( NoError,
      SeqPrim(
        item = IntPrim(),
        evaluated = PJCEvaluated ),
      SeqValued,
      "[ 1,2,3,4 ]",
      [ 1,2,3,4 ] ),
    ( SchemaEvaluationError,
      SeqPrim(
        item = IntPrim(),
        evaluated = PJCEvaluated ),
      SeqValued,
      "[ 1,2,'3',4 ]",
      [ 1,2,3,4 ] ),
    ( NoError,
      MapPrim(
        item = IntPrim(),
        evaluated = PJCEvaluated ),
      MapValued,
      "{ 'a': 1, 'b': 2, 'c': 3 }",
      { 'a': 1, 'b': 2, 'c': 3 } ),
    ( SchemaEvaluationError,
      MapPrim(
        item = IntPrim(),
        evaluated = PJCEvaluated ),
      MapValued,
      "{ 'a': 1, 'b': '2', 'c': 3 }",
      { 'a': 1, 'b': 2, 'c': 3 } ),
    ( NoError,
      MyType1,
      MyType1,
      "{ 'type': 'my_type', 'a': 1 == 1, 'b': 1, 'c': 1.0, 'd': '1', 'e': [ 1, 2, 3 ], 'f': { 'q': 'q', 'r': 'r'} }",
      { 'type': 'my_type', 'a': True, 'b': 1, 'c': 1.0, 'd': '1', 'e': [ 1, 2, 3 ], 'f': { 'q': 'q', 'r': 'r'} } ),
    ( SchemaEvaluationError,
      MyType1,
      MyType1,
      "{ 'type': 'my_type', 'a': 1 == 1, 'b': 1, 'c': 1.0, 'd': '1', 'e': [ 1, 2, '3' ], 'f': { 'q': 'q', 'r': 'r'} }",
      { 'type': 'my_type', 'a': True, 'b': 1, 'c': 1.0, 'd': '1', 'e': [ 1, 2, 3 ], 'f': { 'q': 'q', 'r': 'r'} } ),
    ( SchemaEvaluationError,
      MyType1,
      MyType1,
      "{}",
      {} ),
    ( NoError,
      MyType2,
      MyType2,
      "{ 'type': 'my_type', 'a': 1 == 1, 'b': 1, 'c': 1.0, 'd': '1', 'e': [ 1, 2, 3 ], 'f': { 'q': 'q', 'r': 'r'} }",
      { 'type': 'my_type', 'a': True, 'b': 1, 'c': 1.0, 'd': '1', 'e': [ 1, 2, 3 ], 'f': { 'q': 'q', 'r': 'r'} } ),
    ( SchemaEvaluationError,
      MyType2,
      MyType2,
      "{ 'type': 'my_type', 'a': 1 == 1, 'b': 1, 'c': 1.0, 'd': '1', 'e': [ 1, 2, '3' ], 'f': { 'q': 'q', 'r': 'r'} }",
      { 'type': 'my_type', 'a': True, 'b': 1, 'c': 1.0, 'd': '1', 'e': [ 1, 2, 3 ], 'f': { 'q': 'q', 'r': 'r'} } ),
    ( SchemaEvaluationError,
      MyType2,
      MyType2,
      "{}",
      {} ) ]

  expr_py_tmpl = "$expr:py {}"
  func_py_tmpl = "$func:py\n return {}"

  tmpls = [
    expr_py_tmpl,
    func_py_tmpl ]

  all_cases = list()

  for tmpl in tmpls:
    for case in base_cases:
      exc, schema, cls, src, val = case

      src = tmpl.format(src)

      case = ( exc, schema, cls, src, val )

      all_cases.append( case )

  return all_cases

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
@pytest.fixture( params = _cases() )
def _case( request ):
  return request.param


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_eval_pjc( _case ):

  print( _case )

  exc, schema, cls, src, val = _case

  if exc == NoError:
    ctx = NoRaise()
  else:
    ctx = pytest.raises( exc )

  with ctx:

    f = schema( src )

    print(type(f))
    print(f)

    assert is_evaluated( f ) or is_valued_type( f )
    assert not is_valued( f )

    assert is_evaluated( f ) or isinstance( f, cls )
    assert f._schema is schema.schema


    x = f._eval()

    print( type(x) )
    print(x)


    assert is_valued_type( x )
    assert is_valued( x )

    assert isinstance( x, cls )
    assert x._schema is schema.schema


    assert x == val

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_eval_cmp():
  p = BoolPrim(
    evaluated = PyEvaluated )

  v1 = p('$expr:py x')
  v2 = p('$expr:py x')

  assert v1 == v2
  assert v1 == '$expr:py x'
  assert v1 != '$expr:py y'

  assert v1 != False

  p = IntPrim(
    evaluated = PyEvaluated )

  v1 = p('$expr:py x')
  v2 = p('$expr:py x')

  assert v1 == v2
  assert v1 == '$expr:py x'
  assert v1 != '$expr:py y'

  assert v1 != 0

  p = FloatPrim(
    evaluated = PyEvaluated )

  v1 = p('$expr:py x')
  v2 = p('$expr:py x')

  assert v1 == v2
  assert v1 == '$expr:py x'
  assert v1 != '$expr:py y'

  assert v1 != 0.0

  p = StrPrim(
    evaluated = PyEvaluated )

  v1 = p('$expr:py x')
  v2 = p('$expr:py x')

  assert v1 == v2
  assert v1 == '$expr:py x'
  assert v1 != '$expr:py y'

  assert v1 != ''

  p = MapPrim(
    evaluated = PyEvaluated,
    item = BoolPrim(
      evaluated = PyEvaluated ) )

  v1 = p('$expr:py x')
  v2 = p('$expr:py x')

  assert v1 == v2
  assert v1 == '$expr:py x'
  assert v1 != '$expr:py y'

  assert v1 != {}

  v1 = p({'a' : '$expr:py x'})
  v2 = p({'a' : '$expr:py x'})

  assert v1 == v2
  assert v1 == {'a' : '$expr:py x'}
  assert v1 != {'a' : '$expr:py y'}

  p = SeqPrim(
    evaluated = PyEvaluated,
    item = BoolPrim(
      evaluated = PyEvaluated ) )

  v1 = p('$expr:py x')
  v2 = p('$expr:py x')

  assert v1 == v2
  assert v1 == '$expr:py x'
  assert v1 != '$expr:py y'

  assert v1 != []

  v1 = p(['$expr:py x'])
  v2 = p(['$expr:py x'])

  assert v1 == v2
  assert v1 == ['$expr:py x']
  assert v1 != ['$expr:py y']

  class MyType2( StructValued ):
    schema = dict(
      tag = 'my_type',
      evaluated = PyEvaluated,
      default_val = derived )

    a = BoolPrim(
      evaluated = PyEvaluated,
      default_val = True )

  v1 = MyType2('$expr:py x')
  v2 = MyType2('$expr:py x')

  assert v1 == v2
  assert v1 == '$expr:py x'
  assert v1 != '$expr:py y'

  assert v1 != MyType2.schema.default_val

  v1 = MyType2(a = '$expr:py x')
  v2 = MyType2(a = '$expr:py x')

  assert v1 == v2
  assert v1 == {'type': 'my_type', 'a' : '$expr:py x'}
  assert v1 != {'type': 'my_type', 'a' : '$expr:py y'}

  v1['a'] = '$expr:py y'
  assert v1 == {'type': 'my_type', 'a' : '$expr:py y'}

  assert v1 != MyType2.schema.default_val

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
if __name__ == '__main__':
  test_eval_cmp()



import pytest

from pprint import pprint

import itertools

from partis.schema import (
  required,
  optional,
  derived,
  BoolPrim,
  IntPrim,
  FloatPrim,
  StrPrim,
  SeqPrim,
  MapPrim,
  UnionPrim,
  StructValued,
  PJCEvaluated,
  SchemaError,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  any_schema,
  SchemaDefinitionError )

from partis.schema.serialize.yaml import (
  loads,
  dumps )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_schema():

  # test using metaclass arguments
  class MyType1( StructValued,
    tag = 'my_type',
    struct = [
      ( 'a', BoolPrim() ),
      ( 'b', IntPrim() ),
      ( 'c', FloatPrim() ),
      ( 'd', StrPrim() ),
      ( 'e', SeqPrim(
        item = StrPrim() ) ),
      ( 'f', MapPrim(
        item = StrPrim() ) )  ] ):
    pass

  # test using class namespace arguments
  class MyType2( StructValued ):
    schema = dict(
      tag = 'my_type' )

    a = BoolPrim()
    b = IntPrim()
    c = FloatPrim()
    d = StrPrim()
    e =  SeqPrim(
      item = StrPrim() )

    f = MapPrim(
      item = StrPrim() )

  doc_yaml = """{ type: my_type, a: True, b: 1, c: 1.0, d: "one", e: ["x", "y", "z"], f: { q: q, r: r} }"""

  for cls in [ MyType1, MyType2 ]:
    obj = loads( doc_yaml, cls )

    assert isinstance( obj, cls )
    assert obj._schema is cls.schema
    assert obj._schema.tag == 'my_type'
    assert obj.a == True
    assert obj.b == 1
    assert obj.c == 1.0
    assert obj.d == "one"
    assert obj.e == ["x", "y", "z"]
    assert obj.f == { "q": "q", "r": "r"}

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_struct_proxy():

  # test using metaclass arguments
  class MyType1( StructValued,
    tag = 'my_type',
    struct_proxy = 'a',
    struct = [
      ( 'a', BoolPrim() ),
      ( 'b', IntPrim( default_val = 1 ) ) ] ):
    pass

  x = MyType1( True )

  assert x.a == True
  assert x.b == 1

  with pytest.raises( SchemaDefinitionError ):
    # should raise error because struct_proxy conflicts with union
    u = UnionPrim(
      cases = [
        MyType1,
        BoolPrim() ] )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_struct_no_defaults( ):

  class MyType2( StructValued ):
    schema = dict(
      tag = 'my_type',
      default_val = derived )

    a = BoolPrim(
      default_val = True )

    b = IntPrim(
      default_val = 1234 )

    c = FloatPrim(
      default_val = 1.234 )

    d = StrPrim(
      default_val = 'abcdef' )

  p = MyType2.schema
  print('default_val', p.default_val)

  v = MyType2()

  print(v)
  _v = p.encode( v, no_defaults = True )
  print(_v)
  assert _v == {'type': 'my_type'}

  _v = p.encode( val = { 'a' : True, 'b' : 1234, 'c' : 1.234, 'd' : 'abcdef' }, no_defaults = True )
  print(_v)

  assert _v == {'type': 'my_type'}

  _v = p.encode( val = { 'a' : True, 'b' : 6789, 'c' : 1.234, 'd' : 'xyz' }, no_defaults = True )
  print(_v)

  assert _v == { 'type' : 'my_type', 'b' : 6789, 'd' : 'xyz' }

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_struct_bias( ):

  class MyType( StructValued ):
    schema = dict(
      tag = 'my_type',
      default_val = derived )

    a = BoolPrim(
      default_val = True )

    b = IntPrim(
      default_val = 55,
      min = 1,
      max = 100)

    c = FloatPrim(
      default_val = 55,
      min = 1,
      max = 100)

    d = StrPrim(
      default_val = 'abcd',
      pattern = r'.*',
      nonempty = True,
      max_lines = 3,
      max_cols = 100 )

  v = MyType()
  print(v._bias)
  assert float(v._bias) == 0

  v = MyType(
    a = True,
    b = 10,
    c = 12.34,
    d = 'wxyz' )

  print(v._bias)
  assert float(v._bias) == (
    # keys
    4
    # bool
    + 1
    # int + float
    + 2*3
    # str
    + 8 )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
if __name__ == '__main__':
  test_struct_bias()




import pytest
import sys
from partis.schema import (
  SchemaError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaValidationError,
  SchemaEvaluationError,
  SchemaHint,
  Loc,
  required,
  optional,
  derived,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_optional,
  is_required,
  is_derived,
  is_similar_value_type,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated_class,
  is_evaluated,
  is_valued,
  is_valued_type,
  any_schema,
  BoolValued,
  IntValued,
  FloatValued,
  StrValued,
  SeqValued,
  MapValued,
  PJCEvaluated,
  PassPrim,
  BoolPrim,
  IntPrim,
  FloatPrim,
  StrPrim,
  SeqPrim,
  MapPrim,
  UnionPrim,
  SchemaStruct,
  EvalFunc )

from partis.schema.valued.valued import (
  Valued)

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class NoRaise:
  def __enter__(self):
    return self

  def __exit__(self, type, value, traceback):
    pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NoError = None

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def _cases( ):

  base_types = [
    ( BoolPrim(), BoolValued, [ False, True ] ),
    ( IntPrim(), IntValued, [ -3, -1, 0, 1, 100 ] ),
    ( FloatPrim(), FloatValued, [ -100.1, -3, -1, -3 / 100, 0, 1, 100, 100.1  ] ),
    ( StrPrim(), StrValued, [ 'a', '1', 'True', '0', 'False'  ] ) ]

  all_types = list()

  for type in base_types:
    schema, cls, vals = type

    all_types.append( type )

    all_types.append(
      ( SeqPrim( item = schema ), SeqValued, [ vals, ] ) )

    keys = [
      (chr( 65 + (i % 26) ) + chr( 65 + ((i//26) % 26) ))
      for i in range(len(vals)) ]

    val = dict(list( zip(keys,vals) ))

    all_types.append(
      ( MapPrim( item = schema ), MapValued, [ val, ] ) )

  should_pass_cases = list()
  should_fail_cases = list()

  for type in all_types:
    schema, cls, vals = type

    for val in vals:
      should_pass_cases.append(
        ( NoError, schema, cls, val, val ) )

      # use closure to create evaluable function
      def _f(val):
        def _g( _ ):
          return val

        return _g

      func = _f(val)

      evaluated = PJCEvaluated(
        schema = schema,
        src = EvalFunc( func = func ) )

      should_pass_cases.append(
        ( NoError, schema, cls, evaluated, val ) )

      # src = ( lambda val: (lambda *args, **kwargs: val) )(None)
      #
      # should_fail_cases.append(
      #   ( SchemaEvaluationError, schema, cls, evaluated, val ) )

  all_cases = should_pass_cases + should_fail_cases


  return all_cases

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
@pytest.fixture( params = _cases() )
def _case( request ):
  return request.param


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_valued( _case):

  
  #print( _case )

  exc, schema, cls, input, val = _case

  loc = Loc()

  if exc == NoError:
    ctx = NoRaise()
  else:
    ctx = pytest.raises( exc )

  with ctx:
    x = cls(
      val = input,
      schema = schema,
      loc = loc )

    #print( x._src is input )

    assert is_similar_value_type( x, val )
    assert x._src is input
    assert ( loc is None and isinstance( x._loc, Loc ) ) or x._loc is loc
    assert ( schema is None and isinstance(x._schema, PassPrim) ) or ( x._schema is schema )
    assert ( not is_valued( input ) ) or x == input
    assert is_valued( input ) == is_valued( x )

    _x = x._eval( context = { '_' : None } )

    assert is_similar_value_type( _x, val )
    assert is_valued( _x )
    assert _x == val
    assert _x._loc is x._loc
    assert _x._schema is x._schema

    y = cls( x )

    #print( y._src is input )
    #print( y._src is x )

    assert is_similar_value_type( y, val )
    assert y._src is input
    assert y._loc is x._loc
    assert y._schema is x._schema
    assert is_valued( y ) == is_valued( x )

    _y = y._eval( context = { '_' : None } )

    assert is_similar_value_type( _y, val )
    assert is_valued( _y )
    assert _y == val
    assert _y._loc is y._loc
    assert _y._schema is y._schema


def test_model_hint(_case):
  exc, schema, cls, input, val = _case

  loc = Loc()

  if exc == NoError:
    ctx = NoRaise()
  else:
    ctx = pytest.raises( exc )

  with ctx:
    x = cls(
      val = input,
      schema = schema,
      loc = loc )

    h = x.model_hint()
    assert type(h).__name__ == "ModelHint"


# -*- coding: UTF-8 -*-

import logging
log = logging.getLogger(__name__)

from copy import copy
import inspect
import re

from partis.utils import (
  odict,
  fmt_base_or_type,
  fmt_iterable_or,
  fmt_attr_doc,
  _fmt_class_name )

from partis.schema_meta.base import (
  schema_errors,
  SchemaError,
  SchemaNameError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaHint,
  Bias,
  Loc,
  assert_valid_name,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_declared,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated,
  is_valued,
  is_optional,
  is_required )

from partis.schema_meta.schema import (
  fmt_schema_typename )

from partis.schema.valued import (
  MapValued )

from . import (
  SchemaPrimDeclared,
  SchemaPrim )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class MapPrimDeclared( SchemaPrimDeclared ):
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class MapPrim( SchemaPrim ):
  """Primitive for dictionary/mapping values

  Parameters
  ----------
  item : :class:`SchemaPrim <partis.schema.struct.SchemaStruct>` | :class:`SchemaPrim <partis.schema.prim.base.SchemaPrim>`
    The schema for the value of each item in the mapping
  **kwargs : arguments passed to :class:`SchemaPrim <partis.schema.prim.SchemaPrim>`
  """

  #-----------------------------------------------------------------------------
  def __new__( mcls,
    name,
    bases,
    namespace,
    item,
    min_len = None,
    **kwargs ):

    loc = kwargs.get('loc', None)

    if not is_schema( item ):
      raise SchemaDefinitionError(
        f"`MapPrim` item must be instance of `SchemaPrim` or `Schema`: {item}",
        loc = loc )

    if min_len is not None:
      min_len = int(min_len)

      if min_len <= 0:
        raise SchemaDefinitionError(
          f"`SeqPrim` minimum length must be > 0: {min_len}",
          loc = loc )

    namespace['_p_item'] = item
    namespace['_p_min_len'] = min_len

    if not item.schema_defined:
      kwargs['schema_deps'] = kwargs.get('schema_deps', list()) + [item]

    return super().__new__( mcls,
      name = name,
      bases = bases,
      namespace = namespace,
      valued_type = MapValued,
      **kwargs )

  #-----------------------------------------------------------------------------
  def __init__( cls,
    name,
    bases,
    namespace,
    item,
    min_len = None,
    **kwargs ):


    super().__init__(
      name = name,
      bases = bases,
      namespace = namespace,
      **kwargs )

  #-----------------------------------------------------------------------------
  def _get_attr_doc_lines( cls,
    noindex = False,
    depth = 0,
    max_depth = 2 ):

    lines = super()._get_attr_doc_lines(
      noindex = noindex,
      depth = depth,
      max_depth = max_depth )

    lines.extend([
      fmt_attr_doc(
        name = attr,
        typename = type( getattr( cls, attr ) ),
        val = getattr( cls, attr ),
        prefix = 'schema',
        noindex = noindex )
      for attr in [
        'min_len' ] ])

    lines.append(
      fmt_attr_doc(
        name = 'item',
        typename = fmt_schema_typename( cls.item ),
        val = ...,
        prefix = 'schema',
        noindex = noindex,
        doc = cls.item._get_doc(
          # always specify noindex since the item schema already documented
          noindex = True,
          depth = depth + 1,
          max_depth = max_depth ) ) )

    return lines

  #-----------------------------------------------------------------------------
  def _schema_hash_children( cls ):
    return [ cls.item, ]

  #-----------------------------------------------------------------------------
  @property
  def hints( cls ):
    hints = super().hints

    hints.append( f"min_len: {cls.min_len}" )

    if is_schema_declared(cls.item):
      hints.append( SchemaHint(
        f"item: {cls.item.tag}") )
    else:
      hints.append( SchemaHint(
        f"item:",
        hints = cls.item.schema.hints ) )

    return hints

  #-----------------------------------------------------------------------------
  @property
  def item( cls ):
    return cls._p_item

  #-----------------------------------------------------------------------------
  @property
  def min_len( cls ):
    return cls._p_min_len

  #-----------------------------------------------------------------------------
  def validate( cls,
    val,
    loc = None,
    bias = None ):

    bias = Bias(bias)

    if not is_mapping( val ):
      raise SchemaValidationError(
        f"Must be a `dict`: {type(val).__name__}",
        loc = loc )

    bias += 1

    if cls.min_len:
      bias += 1

      if len(val) < cls.min_len:
        raise SchemaValidationError(
          f"Must have length >= {cls.min_len}: {len(val)}",
          loc = loc )

    return super().validate(
      val = val,
      loc = loc,
      bias = bias )

  #-----------------------------------------------------------------------------
  def encode( cls,
    val = None,
    loc = None,
    no_defaults = None ):

    val, loc, validate = cls._encode(
      val = val,
      loc = loc,
      no_defaults = no_defaults )

    if not validate:
      return val

    val = cls.validate(
      val = val,
      loc = loc )[0]

    item_schema = cls.item.schema.schema

    val = odict([
      ( k, item_schema.encode(
        val = v,
        no_defaults = no_defaults ) )
      for k, v in val.items() ])

    if is_schema_struct( cls.item ):
      # don't need to store tag key since schema is not ambiguous
      tag_key = cls.item.tag_key

      for v in val.values():
        if is_mapping( v ):
          v.pop( tag_key )

    # if no_defaults:
    #   dv = item_schema.default_val
    #
    #   if is_mapping(dv) and is_schema_struct( cls.item ):
    #     dv.pop(tag_key)
    #
    #   equiv = { k: (dv if v is None else v) for k,v in val.items() }
    #
    #   if equiv == cls.default_val:
    #     val = None

    return val

  #-----------------------------------------------------------------------------
  def decode( cls,
    val = None,
    loc = None,
    bias = None ):

    bias = Bias(bias)

    val, loc, validate, bias = cls._decode(
      val = val,
      loc = loc,
      bias = bias )

    if not validate:
      return val

    val, bias = cls.validate(
      val = val,
      loc = loc,
      bias = bias )

    # create new object of the same type as the value (a subclass of dict)
    decode_val = odict()

    for k,v in val.items():
      _loc = loc( val, k )

      with schema_errors(
        msg = f"Dict item could not be decoded",
        data = f"key = '{k}'",
        loc = loc( val, k ),
        hints = [val],
        schema = cls,
        cls = SchemaValidationError ):

        decode_val[k] = cls.item.schema.decode(
          val = v,
          loc = _loc )

      # try:
      #   decode_val[k] = cls.item.schema.decode(
      #     val = v,
      #     loc = _loc )
      #
      # except SchemaError as e:
      #   raise SchemaValidationError(
      #     f"Dict item could not be decoded: {k}",
      #     loc = _loc,
      #     hints = SchemaHint.cast( e ) ) from e

    return cls.valued_type(
      val = decode_val,
      schema = cls,
      loc = loc )
